/**
 * @file   ngram_read_bin.c
 * @author Akinobu LEE
 * @date   Wed Feb 16 17:12:08 2005
 * 
 * <JA>
 * @brief  ХʥN-gramեɤ߹
 *
 * ХʥǤ 2-gram ȵո 3-gram 1ĤΥե
 * ƤޤХʥJuilusȼΤߤ򥵥ݡȤƤꡤ
 * ¾ΥХʥȸߴϤޤΤդƲ
 *
 * rev.3.5 ꡤХʥN-gramΥեΰѹޤ
 * Хȥ Big endian ꤫ޥ¸ѹ(إå
 * Ѵξ򵭽), ޤǥå 24bit 
 *  2-gram ΥХåեǡΰ̤Ԥޤ
 * ˤꡤ3.5 ʹߤ mkbingram ХʥN-gram,
 * 3.4.2 Julius ǤϻȤޤΤդƤ
 * (إååǥ顼Ȥʤ)
 *
 * ʤ 3.5 ʹߤ Julius ǤϽΥǥʤɤ롥ξ,
 * ǥå 24bit ȥХåդΰ̤ϥǥɤ߹߻
 * ٹԤ롥ޤХȥϥإå򸫤ŬѴΤǡ
 * ۤʤХȥΥޥ
 * ХʥN-gramǤʤɤ롥Υǥ⤽Τޤ
 * ɤ߹롥
 * 
 * </JA>
 * 
 * <EN>
 * @brief  Read binary foramt N-gram file
 *
 * In binary format, both 2-gram and reverse 3-gram are stored
 * together in one file.  This binary format is not
 * compatible with other binary format of language model.
 * 
 * From 3.5, internal format of binary N-gram has changed for using
 * machine-dependent natural byte order (previously fixed to big endian),
 * 24bit index and 2-gram backoff compression.  So, binary N-gram
 * generated by mkbingram of 3.5 and later will not work on 3.4.2 and
 * earlier versions.
 *
 * There is full upward- and cross-machine compatibility in 3.5.  Old
 * binary N-gram files still can be read directly, in which case the conversion
 * to 24bit index will performed just after model has been read.
 * Byte order will also considered by header information, so
 * binary N-gram still can be used among different machines.
 * </EN>
 * 
 * $Revision:$
 * 
 */
/*
 * Copyright (c) 1991-2005 Kawahara Lab., Kyoto University
 * Copyright (c) 2000-2005 Shikano Lab., Nara Institute of Science and Technology
 * Copyright (c) 2005      Julius project team, Nagoya Institute of Technology
 * All rights reserved
 */

#include <sent/stddefs.h>
#include <sent/ngram2.h>

static boolean file_is_v4;	///< TRUE if the binary N-gram is made by mkbingram of rev.3.5 and later
static boolean need_swap; ///< TRUE if need byte swap

/** 
 * Binary read function with byte swap (assume file is BIG ENDIAN)
 * 
 * @param fp [in] file pointer
 * @param buf [out] data buffer
 * @param unitbyte [in] unit size in bytes
 * @param unitnum [in] number of unit to read.
 */
static void
rdn(FILE *fp, void *buf, size_t unitbyte, int unitnum)
{
  size_t tmp;
  if ((tmp = myfread(buf, unitbyte, unitnum, fp)) < (size_t)unitnum) {
    perror("ngram_read_bin");
    j_error("read failed\n");
  }
  if (need_swap) {
    if (unitbyte != 1) {
      swap_bytes(buf, unitbyte, unitnum);
    }
  }

}

/** 
 * Check header to see whether the version matches.
 * 
 * @param fp [in] file pointer
 */
static void
check_header(FILE *fp)
{
  char buf[BINGRAM_HDSIZE], *p;
  rdn(fp, buf, 1, BINGRAM_HDSIZE);
  
  p = buf;
  /* version check */
  if (strnmatch(p, BINGRAM_IDSTR, strlen(BINGRAM_IDSTR))) {
    /* bingram file made by mkbingram before 3.4.2 */
    file_is_v4 = FALSE;
    p += strlen(BINGRAM_IDSTR) + 1;
  } else if (strnmatch(p, BINGRAM_IDSTR_V4, strlen(BINGRAM_IDSTR_V4))) {
    /* bingram file made by mkbingram later than 3.5 */
    file_is_v4 = TRUE;
    p += strlen(BINGRAM_IDSTR_V4) + 1;
  } else {
    /* not a bingram file */
    j_printerr("Error: invalid header, you probably use old bingram\n");
    j_error("Error: if so, please re-make with newer mkbingram that comes with Julius-2.0 or later\n");
  }
  /* word size check (for bingram build by mkbingram 3.3p5 and later */
  if (strnmatch(p, BINGRAM_SIZESTR_HEAD, strlen(BINGRAM_SIZESTR_HEAD))) {
    p += strlen(BINGRAM_SIZESTR_HEAD);
    if (! strnmatch(p, BINGRAM_SIZESTR_BODY, strlen(BINGRAM_SIZESTR_BODY))) {
      /* word size does not match (int / short) */
      /*** it should be converted, but just terminate here... ***/
      j_error("Error: word size does not match in bingram\n");
    }
    p += strlen(BINGRAM_SIZESTR_BODY) + 1;

    /* byte order check (v4 (rev.3.5) and later) */
    if (file_is_v4) {
      if (!strnmatch(p, BINGRAM_BYTEORDER_HEAD, strlen(BINGRAM_BYTEORDER_HEAD))) {
	j_error("Error: no information for byte order in v4 format??\n");
      }
      p += strlen(BINGRAM_BYTEORDER_HEAD);
      if (! strnmatch(p, BINGRAM_NATURAL_BYTEORDER, strlen(BINGRAM_NATURAL_BYTEORDER))) {
	/* file endian and running endian is different, need swapping */
	need_swap = TRUE;
      } else {
	need_swap = FALSE;
      }
      p += strlen(BINGRAM_NATURAL_BYTEORDER) + 1;
    }
  } /* if no BINGRAM_SIZESTR_HEAD found, just pass it */

  if (! file_is_v4) {
    /* assume input as big endian */
#ifdef WORDS_BIGENDIAN
    need_swap = FALSE;
#else
    need_swap = TRUE;
#endif
  }
    
  /*j_printf("%s",buf);*/
}

/** 
 * Read a N-gram binary file and store to data.
 * 
 * @param fp [in] file pointer
 * @param ndata [out] N-gram data to store the read data
 * 
 * @return TRUE on success, FALSE on failure.
 */
boolean
ngram_read_bin(FILE *fp, NGRAM_INFO *ndata)
{
  int i,n,len;
  char *w, *p;
  NNID *n3_bgn;
  NNID d, ntmp;
  
  ndata->from_bin = TRUE;

  /* check initial header */
  check_header(fp);
  
  /* read total info and set max_word_num */
  for(n=0;n<MAX_N;n++) {
    rdn(fp, &(ndata->ngram_num[n]), sizeof(NNID), 1);
    if (ndata->ngram_num[n] >= NNIDMAX) {
      j_error("Error: too big %d-gram (%d, should be less than %d)\n", n+1, ndata->ngram_num[n], NNIDMAX);
    }
  }
  ndata->max_word_num = ndata->ngram_num[0];
  if (file_is_v4) rdn(fp, &(ndata->bigram_bo_num), sizeof(NNID), 1);
  
  /* read wname */
  rdn(fp, &len, sizeof(int), 1);
  w = mymalloc(len);
  rdn(fp, w, 1, len);
  /* assign... */
  ndata->wname = (char **)mymalloc(sizeof(char *)*ndata->ngram_num[0]);
  p = w; i = 0;
  while (p < w + len) {
    ndata->wname[i++] = p;
    while(*p != '\0') p++;
    p++;
  }
  if (i != ndata->ngram_num[0]) {
    j_error("wname error??\n");
  }
  /* malloc all */
  ndata->p = (LOGPROB *)mymalloc(sizeof(LOGPROB) * ndata->ngram_num[0]);
  ndata->bo_wt_lr = (LOGPROB *)mymalloc(sizeof(LOGPROB) * ndata->ngram_num[0]);
  ndata->bo_wt_rl = (LOGPROB *)mymalloc(sizeof(LOGPROB) * ndata->ngram_num[0]);
  ndata->n2_bgn = (NNID *)mymalloc(sizeof(NNID) * ndata->ngram_num[0]);
  ndata->n2_num = (WORD_ID *)mymalloc(sizeof(WORD_ID) * ndata->ngram_num[0]);
  ndata->n2tonid = (WORD_ID *)mymalloc(sizeof(WORD_ID) * ndata->ngram_num[1]);
  ndata->p_lr = (LOGPROB *)mymalloc(sizeof(LOGPROB) * ndata->ngram_num[1]);
  ndata->p_rl = (LOGPROB *)mymalloc(sizeof(LOGPROB) * ndata->ngram_num[1]);
  if (file_is_v4) {
    ndata->n2bo_upper = (NNID_UPPER *)mymalloc(sizeof(NNID_UPPER) * ndata->ngram_num[1]);
    ndata->n2bo_lower = (NNID_LOWER *)mymalloc(sizeof(NNID_LOWER) * ndata->ngram_num[1]);
    ndata->bo_wt_rrl = (LOGPROB *)mymalloc(sizeof(LOGPROB) * ndata->bigram_bo_num);
    ndata->n3_bgn_upper = (NNID_UPPER *)mymalloc(sizeof(NNID_UPPER) * ndata->bigram_bo_num);
    ndata->n3_bgn_lower = (NNID_LOWER *)mymalloc(sizeof(NNID_LOWER) * ndata->bigram_bo_num);
    ndata->n3_num = (WORD_ID *)mymalloc(sizeof(WORD_ID) * ndata->bigram_bo_num);
  } else {
    ndata->bo_wt_rrl = (LOGPROB *)mymalloc(sizeof(LOGPROB) * ndata->ngram_num[1]);
    ndata->n3_bgn_upper = (NNID_UPPER *)mymalloc(sizeof(NNID_UPPER) * ndata->ngram_num[1]);
    ndata->n3_bgn_lower = (NNID_LOWER *)mymalloc(sizeof(NNID_LOWER) * ndata->ngram_num[1]);
    ndata->n3_num = (WORD_ID *)mymalloc(sizeof(WORD_ID) * ndata->ngram_num[1]);
  }
  ndata->n3tonid = (WORD_ID *)mymalloc(sizeof(WORD_ID) * ndata->ngram_num[2]);
  ndata->p_rrl = (LOGPROB *)mymalloc(sizeof(LOGPROB) * ndata->ngram_num[2]);

  if (! file_is_v4) {
    n3_bgn = (NNID *)mymalloc(sizeof(NNID) * ndata->ngram_num[1]);
  }
  
  /* malloc work area */

  /* read 1-gram */
  j_printerr("1-gram.");
  rdn(fp, ndata->p, sizeof(LOGPROB), ndata->ngram_num[0]);
  j_printerr(".");
  rdn(fp, ndata->bo_wt_lr, sizeof(LOGPROB), ndata->ngram_num[0]);
  j_printerr(".");
  rdn(fp, ndata->bo_wt_rl, sizeof(LOGPROB), ndata->ngram_num[0]);
  j_printerr(".");
  rdn(fp, ndata->n2_bgn, sizeof(NNID), ndata->ngram_num[0]);
  j_printerr(".");
  rdn(fp, ndata->n2_num, sizeof(WORD_ID), ndata->ngram_num[0]);
  
  /* read 2-gram*/
  j_printerr("2-gram.");
  rdn(fp, ndata->n2tonid, sizeof(WORD_ID), ndata->ngram_num[1]);
  j_printerr(".");
  rdn(fp, ndata->p_lr, sizeof(LOGPROB), ndata->ngram_num[1]);
  j_printerr(".");
  rdn(fp, ndata->p_rl, sizeof(LOGPROB), ndata->ngram_num[1]);
  j_printerr(".");
  if (file_is_v4) {
    rdn(fp, ndata->n2bo_upper, sizeof(NNID_UPPER), ndata->ngram_num[1]);
    rdn(fp, ndata->n2bo_lower, sizeof(NNID_LOWER), ndata->ngram_num[1]);
    j_printerr(".");
    rdn(fp, ndata->bo_wt_rrl, sizeof(LOGPROB), ndata->bigram_bo_num);
    j_printerr(".");
    rdn(fp, ndata->n3_bgn_upper, sizeof(NNID_UPPER), ndata->bigram_bo_num);
    rdn(fp, ndata->n3_bgn_lower, sizeof(NNID_LOWER), ndata->bigram_bo_num);
    j_printerr(".");
    rdn(fp, ndata->n3_num, sizeof(WORD_ID), ndata->bigram_bo_num);
  } else {
    rdn(fp, ndata->bo_wt_rrl, sizeof(LOGPROB), ndata->ngram_num[1]);
    j_printerr(".");
    rdn(fp, n3_bgn, sizeof(NNID), ndata->ngram_num[1]);
    for(d=0;d<ndata->ngram_num[1];d++) {
      if (n3_bgn[d] == NNID_INVALID) {
	ndata->n3_bgn_lower[d] = 0;
	ndata->n3_bgn_upper[d] = NNID_INVALID_UPPER;
      } else {
	ntmp = n3_bgn[d] & 0xffff;
	ndata->n3_bgn_lower[d] = ntmp;
	ntmp = n3_bgn[d] >> 16;
	ndata->n3_bgn_upper[d] = ntmp;
      }
    }
    j_printerr(".");
    rdn(fp, ndata->n3_num, sizeof(WORD_ID), ndata->ngram_num[1]);
  }

  /* read 3-gram*/
  j_printerr("3-gram.");
  rdn(fp, ndata->n3tonid, sizeof(WORD_ID), ndata->ngram_num[2]);
  j_printerr(".");
  rdn(fp, ndata->p_rrl, sizeof(LOGPROB), ndata->ngram_num[2]);

  /* make word search tree for later lookup */
  j_printerr("indexing...");
  ngram_make_lookup_tree(ndata);

  /* compact the 2-gram back-off and 3-gram links */
  if (! file_is_v4) {
    free(n3_bgn);
    ngram_compact_bigram_context(ndata);
  }
  
  /* set unknown id */
  set_unknown_id(ndata);
  
  return TRUE;
}
